
from types import ModuleType
from typing import TYPE_CHECKING

try: 
    import fastapi
    from fastapi import FastAPI
    _fastapi_available = True

except ImportError: 
    fastapi: ModuleType = None
    FastAPI: object = None
    _fastapi_available = False

try: 
    import dateparser
    _dateparser_available = True
except ImportError: 
    dateparser: ModuleType = None
    _dateparser_available = False

try: 
    import pytz
    _pytz_available = True
except ImportError: 
    pytz: ModuleType = None
    _pytz_available = False

try: 
    import starlette
    from starlette.requests import Request as _Request
    _starlette_available = True
except ImportError: 
    starlette: ModuleType = None
    _Request: object = None
    _starlette_available = False

try: 
    import httpx
    from httpx import Response as HttpResponse
    from httpx import Request as HttpRequest
    _httpx_available = True
except ImportError: 
    httpx: ModuleType = None
    HttpRequest: object = None
    HttpResponse: object = None
    _httpx_available = False

from lazy.libz import Lib

_LAZYAPI_CHECKED = False


def _ensure_api_reqs():
    """
    Checks to ensure that all 3rd party libs required by this library
    are met. 
    """
    global _LAZYAPI_CHECKED
    if _LAZYAPI_CHECKED: return
    global fastapi, starlette, httpx, dateparser, pytz
    global _Request, HttpRequest, HttpResponse, FastAPI
    
    if fastapi is None:
        fastapi = Lib.import_lib('fastapi', 'fastapi[all]')
        Lib.reload_module(fastapi)
        FastAPI = fastapi.FastAPI
    if starlette is None:
        starlette = Lib.import_lib('starlette')
        Lib.reload_module(starlette)
        _Request = starlette.requests.Request
    if httpx is None:
        httpx = Lib.import_lib('httpx')
        Lib.reload_module(httpx)
        HttpResponse = httpx.Response
        HttpRequest = httpx.Request
    if dateparser is None:
        dateparser = Lib.import_lib('httpx')
        Lib.reload_module(dateparser)
    if pytz is None:
        pytz = Lib.import_lib('pytz')
        Lib.reload_module(pytz)
    _LAZYAPI_CHECKED = True

__all__ = (
    'fastapi',
    'starlette',
    'httpx',
    'pytz',
    'dateparser',
    '_ensure_api_reqs',
    '_Request', 
    'HttpRequest', 
    'HttpResponse',
    'TYPE_CHECKING',
    'Lib',
    '_starlette_available',
    '_fastapi_available',
    '_httpx_available',
    '_pytz_available',
    '_dateparser_available',
)

    


from __future__ import annotations

import os

from .base import *
from .cfs_base import Minio_CFS
from .cfs_pathz_base import *


class PathzMinioPurePath(PathzCFSPurePath):
    _prefix: str = 'minio'
    _posix_prefix: str = 's3'
    _provider: str = 'Minio'
    _win_pathz: ClassVar = 'PurePathzMinioWindowsPath'
    _posix_pathz: ClassVar = 'PurePathzMinioPosixPath'


class PurePathzMinioPosixPath(PurePathzCFSPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PurePathzMinioWindowsPath(PurePathzCFSWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()

class PathzMinioPath(PathzCFSPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 'minio'
    _posix_prefix = 's3'
    _provider = 'Minio'

    _win_pathz: ModuleType = 'PathzMinioWindowsPath'
    _posix_pathz: ModuleType = 'PathzMinioPosixPath'

    def _init(self, template: Optional['PathzMinioPath'] = None):
        self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is PathzMinioPath or issubclass(cls, PathzMinioPath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self

    @property
    def _cloudpath(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        if self._prefix in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        return self._pathlike.join(*self.parts)

class PathzMinioPosixPath(PosixPath, PathzMinioPath, PurePathzMinioPosixPath):
    __slots__ = ()


class PathzMinioWindowsPath(WindowsPath, PathzMinioPath, PurePathzMinioWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("PathzMinioPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("PathzMinioPath.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        PathzMinioPurePath, PathzMinioPath, PurePathzMinioPosixPath, PathzMinioWindowsPath, PathzMinioPosixPath, PurePathzMinioWindowsPath
    ]
)

MinioFileSystem = Minio_CFS


__all__ = (
    'PathzMinioPurePath',
    'PathzMinioPath',
    'PurePathzMinioPosixPath',
    'PathzMinioWindowsPath',
    'PathzMinioPosixPath',
    'PurePathzMinioWindowsPath',
    'MinioFileSystem'
)



import orjson as _orjson
from pydantic import BaseModel, Field
from typing import List, Any
from .utils import to_camelcase


class OrJson:
    binary: False

    @classmethod
    def dumps(cls, obj, *args, default: Any = None, **kwargs):
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()

    @classmethod
    def loads(cls, obj, *args, **kwargs):
        return _orjson.loads(obj, *args, **kwargs)

class BaseCls(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = 'allow'
        json_loads = OrJson.loads
        json_dumps = OrJson.dumps

    
    def get(self, name, default: Any = None):
        return getattr(self, name, default)

class BaseLazy(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = 'allow'
        alias_generator = to_camelcase
        json_loads = OrJson.loads
        json_dumps = OrJson.dumps

    def get(self, name, default: Any = None):
        return getattr(self, name, default)

class Validator(BaseCls):
    text: str
    include: List[str] = []
    exclude: List[str] = []
    exact: bool = False

class ValidatorArgs(BaseCls):
    include: List[str] = []
    exclude: List[str] = []


class BaseDataCls(BaseCls):
    string: str = None
    value: Any = None
    dtype: str = None

__all__ = [
    'BaseModel',
    'Field',
    'BaseCls',
    'BaseDataCls',
    'BaseLazy'
]


import dill

from lazy.types import *
from lazy.types.pyd import ByteSize
from lazy.models import BaseCls
from lazy.configz import ConfigCls
from lazy.utils import get_logger

# https://phiresky.github.io/blog/2020/sqlite-performance-tuning/
# https://blog.devart.com/increasing-sqlite-performance.html
# page_size: 4096 seems to be the fastest-ish

"""
You can reload the settings manually at runtime

from lazy.io.cachez import CachezConfigz

CachezConfigz.update_config(sqlmode = 'perf', serializer = 'pickle')
print(CachezConfigz.dict())

"""

logger = get_logger('cachez')

class SqlConfig(BaseCls):
    mode: str = ''
    settings: Dict[str, Any] = {}
    policies: Dict[str, Any] = {}

class CachezConfigz(ConfigCls):
    dbname: str = 'cache.db'
    default_name: str = 'cachez'
    sqlmode: str = 'default' # default / standard / perf / optimized
    default_table: str = 'Cache'
    serializer: str = 'dill' # dill / pickle
    dataset_mode: bool = False # if enabled, will start Index at 0 rather than 500 trill

    default_compression_level: int = 3
    standard_compression_level: int = 5
    optim_compression_level: int = 7
    perf_compression_level: int = 0 # Zero compression

    default_size_limit: ByteSize = '1 GiB'
    standard_size_limit: ByteSize = '5 GiB'
    optim_size_limit: ByteSize = '10 GiB'
    perf_size_limit: ByteSize = '20 GiB'

    default_disk_min_file_size: ByteSize = '32 KiB'
    standard_disk_min_file_size: ByteSize = '64 KiB'
    optim_disk_min_file_size: ByteSize = '128 KiB'
    perf_disk_min_file_size: ByteSize = '256 KiB'
    
    default_sqlite_mmap_size: ByteSize = '512 MiB'
    standard_sqlite_mmap_size: ByteSize = '1024 MiB'
    optim_sqlite_mmap_size: ByteSize = '1024 MiB'
    perf_sqlite_mmap_size: ByteSize = '2 GiB'
    
    default_sqlite_cache_size: int = '4096' # '8192' # 8,192 pages
    standard_sqlite_cache_size: int = '4096' # '16384' # 16,384 pages
    optim_sqlite_cache_size: int = '4096' # '8192' # 8,192 pages
    perf_sqlite_cache_size: int = '4096' # '32768' # 32,768 pages

    def get(self, key: str, default=None):
        if hasattr(self, key): return getattr(self, key, default)
        if hasattr(self, f'{self.sqlmode}_{key}'): return getattr(self, f'{self.sqlmode}_{key}', default)
        return default
    
    @property
    def start_index_n(self):
        if not self.dataset_mode: return 500000000000000
        logger.info('Dataset mode is Enabled. Index will start at 0')
        return 0

    @property
    def default_keys(self):
        return [u'statistics', u'tag_index', u'eviction_policy', u'cull_limit', u'sqlite_auto_vacuum', u'sqlite_cache_size', u'sqlite_journal_mode', u'sqlite_mmap_size', u'sqlite_synchronous', u'disk_min_file_size', u'disk_pickle_protocol']

    def get_default_sql_settings(self):
        return {
            u'statistics': 0,  # False
            u'tag_index': 0,  # False
            u'eviction_policy': u'least-recently-stored',
            u'size_limit': self.default_size_limit,
            u'cull_limit': 10,
            u'sqlite_auto_vacuum': 1,  # FULL
            u'sqlite_cache_size': self.default_sqlite_cache_size,
            u'sqlite_journal_mode': u'wal',
            u'sqlite_mmap_size': self.default_sqlite_mmap_size,
            u'sqlite_synchronous': 1,  # NORMAL
            u'disk_min_file_size': self.default_disk_min_file_size,  # 32kb
            u'disk_pickle_protocol': dill.DEFAULT_PROTOCOL,
        }
    
    def get_standard_sql_settings(self):
        return {
            u'statistics': 0,  # False
            u'tag_index': 0,  # False
            u'eviction_policy': u'least-recently-stored',
            u'size_limit': self.standard_size_limit,
            u'cull_limit': 10,
            u'sqlite_auto_vacuum': 1,  # FULL
            u'sqlite_cache_size': self.standard_sqlite_cache_size,
            u'sqlite_journal_mode': u'wal',
            u'sqlite_mmap_size': self.standard_sqlite_mmap_size,
            u'sqlite_synchronous': 1,  # NORMAL
            u'disk_min_file_size': self.standard_disk_min_file_size,  # 32kb
            u'disk_pickle_protocol': dill.HIGHEST_PROTOCOL,
        }
    
    def get_optim_sql_settings(self):
        return {
            u'statistics': 0,  # False
            u'tag_index': 0,  # False
            u'eviction_policy': u'least-recently-stored',
            u'size_limit': self.optim_size_limit,
            u'cull_limit': 10,
            u'sqlite_auto_vacuum': 1,  # FULL
            u'sqlite_cache_size': self.optim_sqlite_cache_size,
            u'sqlite_journal_mode': u'wal',
            u'sqlite_mmap_size': self.optim_sqlite_mmap_size,
            u'sqlite_synchronous': 1,  # NORMAL
            u'disk_min_file_size': self.optim_disk_min_file_size,  # 32kb
            u'disk_pickle_protocol': dill.HIGHEST_PROTOCOL,
        }

    def get_perf_sql_settings(self):
        return {
            u'statistics': 0,  # False
            u'tag_index': 0,  # False
            u'eviction_policy': u'least-frequently-used',
            u'size_limit': self.perf_size_limit,
            u'cull_limit': 10,
            u'sqlite_auto_vacuum': 1,  # FULL
            u'sqlite_cache_size': self.perf_sqlite_cache_size,
            u'sqlite_journal_mode': u'wal',
            u'sqlite_mmap_size': self.perf_sqlite_mmap_size,
            u'sqlite_synchronous': 0,  # OFF
            u'disk_min_file_size': self.perf_disk_min_file_size,  # 32kb
            u'disk_pickle_protocol': dill.HIGHEST_PROTOCOL,
        }

    @property
    def sql_settings(self):
        if self.sqlmode in {'perf', 'max', 'performance'}:
            return self.get_perf_sql_settings()
        if self.sqlmode in {'std', 'standard', 'normal'}:
            return self.get_standard_sql_settings()
        if self.sqlmode in {'opt', 'optimized', 'optim', 'optimize'}:
            return self.get_optim_sql_settings()
        return self.get_default_sql_settings()


    @property
    def compression_lvl(self):
        if self.sqlmode in {'perf', 'max', 'performance'}:
            return self.perf_compression_level
        if self.sqlmode in {'std', 'standard', 'normal'}:
            return self.standard_compression_level
        if self.sqlmode in {'opt', 'optimized', 'optim', 'optimize'}:
            return self.optim_compression_level
        return self.default_compression_level


    def get_eviction_policies(self, table_name: str = None):
        table_name = table_name or self.default_table
        return {
            'none': {
                'init': None,
                'get': None,
                'cull': None,
            },
            'least-recently-stored': {
                'init': (
                    f'CREATE INDEX IF NOT EXISTS {table_name}_store_time ON'
                    f' {table_name} (store_time)'
                ),
                'get': None,
                'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY store_time LIMIT ?',
            },
            'least-recently-used': {
                'init': (
                    f'CREATE INDEX IF NOT EXISTS {table_name}_access_time ON'
                    f' {table_name} (access_time)'
                ),
                'get': 'access_time = {now}',
                'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY access_time LIMIT ?',
            },
            'least-frequently-used': {
                'init': (
                    f'CREATE INDEX IF NOT EXISTS {table_name}_access_count ON'
                    f' {table_name} (access_count)'
                ),
                'get': 'access_count = access_count + 1',
                'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY access_count LIMIT ?',
            },
        }


    def get_sql_settings(self, table_name: str = None, **config) -> SqlConfig:
        settings = self.sql_settings
        if config: settings.update(config)
        policies = self.get_eviction_policies(table_name = table_name)
        return SqlConfig(mode = self.sqlmode, settings = settings, policies = policies)





import anyio
import inspect
from collections import deque
from datetime import datetime, timedelta
from anyio.to_thread import run_sync as _run_sync
from functools import wraps, partial, lru_cache
from typing import Callable, Dict, List, Any, Union, Iterable, Generator, Awaitable, Optional, Coroutine

try:
    from tqdm.auto import tqdm, trange
    from tqdm.asyncio import trange as async_trange
    from tqdm.asyncio import tqdm as async_tqdm

    _tqdm_enabled = True
except ImportError:
    tqdm: object = None
    trange: object = None
    async_trange: object = None
    async_tqdm: object = None
    _tqdm_enabled = False

try:
    import trio
    _async_backend = 'trio'

except ImportError:
    _async_backend = 'asyncio'

from .loops import *


CoroutineResult = Awaitable[Any]
CoroutineFunction = Callable[..., CoroutineResult]
CoroutineMethod = Callable[..., CoroutineResult]


def iscoroutinefunction(obj):
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    if inspect.iscoroutinefunction(obj): return True
    if hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__): return True
    return False


async def to_thread(func: Callable, *args, **kwargs) -> Any:
    # anyio's run_sync() doesn't support passing kwargs
    func_kwargs = partial(func, **kwargs)
    return await _run_sync(func_kwargs, *args)


async def run_sync(func: Callable, *args, **kwargs) -> Any:
    # anyio's run_sync() doesn't support passing kwargs
    func_kwargs = partial(func, **kwargs)
    return await _run_sync(func_kwargs, *args)



def run_async_asyncio(func, *args, **kwargs):
    """
    Runs an Async Function in a Sync Call using asyncio
    in sync calls
    """
    if not iscoroutinefunction(func):
        func = AsyncFunction(func)
    coro = func(*args, **kwargs)
    return asyncio_run(coro)


def run_async_anyio(func, *args, **kwargs):
    """
    Runs an Async Function in a Sync Call using anyio
    Returns the Results
    """
    current_async_module = getattr(anyio._core._eventloop.threadlocals, "current_async_module", None)
    partial_func = partial(func, *args, **kwargs)
    try:
        if current_async_module is None:
            return anyio.run(partial_func)
            #return anyio.run(partial_func, backend = _async_backend)
        return anyio.from_thread.run(partial_func)
    except:
        return run_async_asyncio(func, *args, **kwargs)


def run_async(func: Coroutine, *args, **kwargs):
    """
    Returns the result of the coroutine using asyncio/anyio
    in sync calls
    """
    if _LOOP_ACTIVE(): 
        return run_async_asyncio(func, *args, **kwargs)
    return run_async_anyio(func, *args, **kwargs)

"""
Utilities
"""

def count_iterable(iterable):
    if hasattr(iterable, '__len__'):
        return len(iterable)
    d = deque(enumerate(iterable, 1), maxlen=1)
    return d[0][0] if d else 0

async def async_count_iterable(iterable):
    if hasattr(iterable, '__len__'):
        return len(iterable)
    #return sum(1 for _ in await iterable())
    d = deque(enumerate(await iterable(), 1), maxlen=1)
    return d[0][0] if d else 0

def wrap_tqdm_func(func: Callable, desc: str = None, **config):
    desc = desc + f' {func.__name__}' if desc else func.__name__
    @wraps(func)
    def wrapped_tqdm_func(*args, **kwargs) -> 'trange':
        func_iter = partial(func, *args, **kwargs)
        return trange(count_iterable(func_iter()), desc = desc,  **config)
    return wrapped_tqdm_func
    

def wrap_tqdm_iterable(func: Callable, desc: str = None, leave: bool = False, **config) -> trange:
    """
    Assumes that the func is already wrapped.
    May have some overhead.
    """
    #_func = sync_to_async_func(func) if not iscoroutinefunction(func) else func
    #_func_length = count_iterable(anyio.run_async_from_thread(func)) if iscoroutinefunction(func) else count_iterable(get_async_run_func(func))
    #return trange(_func_length, desc = desc, leave = leave, **config)
    return trange(count_iterable(func()), desc = desc, leave = leave, **config)

"""
Wraps
"""


def timed_cache(seconds: int, maxsize: int = 128):
    def wrapper_cache(func: Callable):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.utcnow() + func.lifetime
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if datetime.utcnow() >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.utcnow() + func.lifetime
            return func(*args, **kwargs)
        return wrapped_func
    return wrapper_cache


def AsyncFunction(func: Callable) -> CoroutineFunction:
    """
    Converts a Sync Function to Async Function
    """
    @wraps(func)
    async def new_func(*args, **kwargs) -> Any:
        return await run_sync(func, *args, **kwargs)
    return new_func

def AsyncClassFunction(func: Callable) -> CoroutineFunction:
    """
    Converts a Sync Class Function to Async Class Function
    """
    @wraps(func)
    async def new_func(self, *args, **kwargs) -> Any:
        return await run_sync(func, *args, **kwargs)
    return new_func


def AsyncMethod(func: Callable) -> CoroutineMethod:
    """
    Converts a Sync Class Method to Async Class Function
    """
    @wraps(func)
    async def method(self, *args, **kwargs) -> Any:
        return await run_sync(func, *args, **kwargs)
    return method


def AsyncCoroMethod(coro: CoroutineFunction) -> CoroutineMethod:
    """
    Converts an Async Coroutine to Async Class Function
    """
    @wraps(coro)
    async def method(self, *args, **kwargs) -> Any:
        return await coro(*args, **kwargs)
    return method


def SyncFunction(func: CoroutineFunction):
    """
    Converts an Async Function to a Sync Function
    """
    @wraps(func)
    def wrapped_func(*args, **kwargs) -> Any:
        if _LOOP_ACTIVE(): 
            return run_async_asyncio(func, *args, **kwargs)
        return run_async_anyio(func, *args, **kwargs)
    return wrapped_func




__all__ = (
    'run_sync',
    'to_thread',
    'run_async',
    'run_async_anyio',
    'run_async_asyncio',
    'asyncio_run',
    'asyncio_gather',
    'SyncFunction',
    'AsyncFunction',
    'AsyncClassFunction',
    'AsyncMethod',

    'AsyncCoroMethod',
    'CoroutineResult',
    'CoroutineFunction',
    'CoroutineMethod',
    'tqdm',
    'trange',
    '_tqdm_enabled',
    '_async_backend',
    'iscoroutinefunction',
    'wrap_tqdm_iterable',
    
)




"""
Taken from 
https://stackoverflow.com/questions/52232177/runtimeerror-timeout-context-manager-should-be-used-inside-a-task/69514930#69514930

"""

import asyncio
import threading
from typing import Awaitable, TypeVar

T = TypeVar("T")

def _start_background_loop(loop):
    asyncio.set_event_loop(loop)
    loop.run_forever()


_LOOP: asyncio.AbstractEventLoop = None
_LOOP_THREAD: threading.Thread = None

_LOOP_ACTIVE = lambda: bool(_LOOP is not None)

def _ensure_loop_started():
    """
    Used to Start loop when needed.
    """
    global _LOOP, _LOOP_THREAD
    if _LOOP is not None: return
    _LOOP = asyncio.new_event_loop()
    _LOOP_THREAD = threading.Thread(target=_start_background_loop, args=(_LOOP,), daemon=True)
    _LOOP_THREAD.start()

def get_loop():
    """
    Returns the created EventLoop.
    Starts it if it isnt running.
    """
    _ensure_loop_started()
    return _LOOP


def asyncio_run(coro: Awaitable[T], timeout=30) -> T:
    """
    Runs the coroutine in an event loop running on a background thread,
    and blocks the current thread until it returns a result.
    This plays well with gevent, since it can yield on the Future result call.

    :param coro: A coroutine, typically an async method
    :param timeout: How many seconds we should wait for a result before raising an error
    """
    _ensure_loop_started()
    return asyncio.run_coroutine_threadsafe(coro, _LOOP).result(timeout=timeout)


def asyncio_gather(*futures, return_exceptions=False):
    """
    A version of asyncio.gather that runs on the internal event loop
    """
    _ensure_loop_started()
    return asyncio.gather(*futures, loop=_LOOP, return_exceptions=return_exceptions)


__all__ = (
    'get_loop',
    'asyncio_run',
    'asyncio_gather',
    '_LOOP_ACTIVE'
)



from lazy.types import *
from lazy.models.base import BaseCls, Field, OrJson, to_camelcase
from lazy.serialize import Serializer
from lazy.serialize._pysimd import SimdArray, SimdObject
from lazy.utils.wrapz.utils import timed_cache


from typing import overload
from .base_imports import *
from .utils import convert_to_cls
from .timez import get_timestamp_utc

if _httpx_available:
    from httpx import Response as HttpReponse
    from httpx import Request as HttpRequest
else:
    HttpReponse, HttpRequest = object, object

if _starlette_available:
    from starlette.requests import Request as _Request
else:
    _Request = object

RequestType = TypeVar('RequestType', Dict[Any, Any], HttpRequest, _Request)
JsonResponse = TypeVar('JsonResponse', HttpReponse, Union[None, int], None, Union[HttpReponse, int], Union[Dict[Any, Any], int], Dict[Any, Any], Dict[str, Any], SimdObject)
TextResponse = TypeVar('TextResponse', HttpReponse, Union[None, int], None, Union[HttpReponse, int], Union[str, int], str)
ContentResponse = TypeVar('ContentResponse', HttpReponse, Union[None, int], None, Union[HttpReponse, int], Union[bytes, int], Union[str, int], bytes, str)
AnyResponse = TypeVar('AnyResponse', HttpReponse, Union[None, int], None, Union[HttpReponse, int], Union[Any, int], Any, bytes, str, SimdObject, SimdArray)
BoolResponse = TypeVar('BoolResponse', HttpReponse, Union[bool, int], bool, Union[HttpReponse, int], Union[bool, int], bool)

ContentType = TypeVar('ContentType', str, bytes)
DataType = TypeVar('DataType', Dict[str, Any], Dict[Any, Any], List[Any], str)
DataObjType = TypeVar('DataObjType', SimdArray, SimdObject, Dict[str, Any], Dict[Any, Any], List[Any])


class Response(BaseCls):
    resp: HttpReponse
    client_type: str = 'sync'
    method: str = 'get'
    timestamp: str = Field(default_factory=get_timestamp_utc)

    class Config:
        arbitrary_types_allowed = True
        extra = 'allow'
        alias_generator = to_camelcase
        json_loads = OrJson.loads
        json_dumps = OrJson.dumps
    
    @overload
    def __getattr__(self, name, default = None):
        if not hasattr(self, name, None):
            return getattr(self.resp, name, default)
        return getattr(self, name, default)


    @property
    def status_code(self): return self.resp.status_code
    @property
    def text(self) -> str: return self.resp.text
    @property
    def content(self) -> ContentType: return self.resp.content
    @property
    @timed_cache(10)
    def data(self) -> DataType: return self.resp.json()
    @property
    def data_obj(self) -> DataObjType: return Serializer.SimdJson.parse(self.resp)
    @property
    def url(self): return self.resp.url
    @property
    def is_async(self): return bool(self.client_type == 'async')
    @property
    def is_sync(self): return bool(self.client_type == 'sync')
    @property
    def is_error(self): return bool(self.status_code >= 400)
    @property
    def is_success(self): return bool(self.status_code < 300)
    @property
    def is_redirect(self): return bool(self.status_code in range(300, 399))
    @property
    def status(self): return self.resp.status_code
    
    @property
    @timed_cache(10)
    def _valid_data(self):
        try: return self.data
        except: return None

    @property
    def data_cls(self) -> Type[BaseCls]:
        """Attempts to turn the response (if valid json/dict) to lazycls object. Returns None if not valid"""
        if not self._valid_data: return None
        return convert_to_cls(resp=self._valid_data)
    
    @property
    def data_yaml(self) -> Optional[Union[List[Any], Dict[Any, Any]]]:
        """ Attempts to transform into valid YAML """
        return Serializer.Yaml.loads(self.text)

    @property
    def data_b64(self) -> str:
        """ Attempts to Serialize the Data into Base64(str)"""
        return Serializer.Base.b64_encode(self.text)

    @property
    def data_bgz(self) -> str:
        """ Attempts to Serialize the Data into Base64(Gzip(str))"""
        return Serializer.Base.b64_gzip_encode(self.text)
    
    @property
    def data_pickle(self) -> bytes:
        """ Attempts to Serialize the Data into Picke(content)"""
        return Serializer.Pkl.dumps(self.content)
    
    @property
    def data_b64_decode(self) -> str:
        """ Attempts to Deserialize the Data from Base64(bytes)"""
        return Serializer.Base.b64_decode(self.content)

    @property
    def data_bgz_decode(self) -> str:
        """ Attempts to Deserialize the Data from Base64(Gzip(str))"""
        return Serializer.Base.b64_gzip_decode(self.content)
    
    @property
    def data_picke_decode(self) -> Any:
        """ Attempts to Deserialize the Data from Picke(content)"""
        return Serializer.Pkl.loads(self.content)
    



__all__ = (
    'Response',
    'HttpResponse',
    'HttpRequest',
    'RequestType',
    'JsonResponse',
    'TextResponse',
    'ContentResponse',
    'AnyResponse',
    'BoolResponse',
    'HttpResponse',
    'ContentType',
    'DataType',
    'DataObjType'
)

    






from .loops import get_loop, asyncio_run, asyncio_gather, _LOOP_ACTIVE
from .utils import (
    run_sync,
    to_thread,
    run_async,
    run_async_anyio,
    run_async_asyncio,
    asyncio_run,
    asyncio_gather,
    SyncFunction,
    AsyncFunction,
    AsyncClassFunction,
    AsyncMethod,
    AsyncCoroMethod,
    CoroutineResult,
    CoroutineFunction,
    CoroutineMethod,
    iscoroutinefunction
    )

from .compute import (
    AsyncTaskGroup,
    ProcessPool,
    get_mp
)

from .retry import retryable, async_retryable


from __future__ import annotations

import time
from datetime import datetime, timezone
from lazy.models.timez import TimeCls
from .config import TimeZoneConfigz

from .base_imports import *

def _require_pytz():
    pytz = Lib.import_lib('pytz')
    Lib.reload_module(pytz)

def _require_dateparser():
    dateparser = Lib.import_lib('dateparser')
    Lib.reload_module(dateparser)

if _pytz_available:
    import pytz
    api_timezone = pytz.timezone(TimeZoneConfigz.desired)
    utc_timezone = pytz.timezone("UTC")
    dtime_now_tz = lambda: datetime.now(api_timezone)
else:
    api_timezone = _require_pytz
    utc_timezone = _require_pytz
    dtime_now_tz = _require_pytz

if _dateparser_available:
    import dateparser

dtime_now = lambda: datetime.now()
dtime_now_utc = lambda: datetime.now(timezone.utc)

def timer(s: float = None):
    return time.time() if not s else (time.time() - s)


def dtime_parse(timeframe: str = '30 mins', future: bool = False):
    if not _dateparser_available: _require_dateparser()
    if future:
        timeframe = 'in ' + timeframe
        prefer = 'future'
    else:
        timeframe += ' ago'
        prefer = 'past'
    return dateparser.parse(timeframe, settings={'PREFER_DATES_FROM': prefer, 'TIMEZONE': 'UTC', 'RETURN_AS_TIMEZONE_AWARE': True})


def dtime_diff(dtime: datetime = None, timeframe: str = '30 mins', future: bool = False, secs_only: bool = False):
    if not dtime: dtime = datetime.now(timezone.utc)
    dtime_to_obj = dtime_parse(timeframe=timeframe, future=future)
    dtime_diff = (dtime_to_obj - dtime) if future else (dtime - dtime_to_obj)
    if secs_only: return dtime_diff.total_seconds()
    return dtime_diff

def get_dtime(dtime: datetime = None, start: datetime = None, timeframe: str = None, future: bool = False) -> datetime:
    if dtime and start: return start - dtime
    if dtime and timeframe: return dtime_diff(dtobj=dtime, timeframe=timeframe, future=future)
    if timeframe: return dtime_parse(timeframe=timeframe, future=future)
    return dtime_now_utc()

def get_dtime_str(dtime: datetime = None, start: datetime = None, timeframe: str = None, future: bool = False, tz_format: bool = True, dt_format: str = None):
    dt = get_dtime(dtime=dtime, start=start, timeframe=timeframe, future=future)
    if tz_format: return dt.strftime(TimeZoneConfigz.tz_format)
    if dt_format: return dt.strftime(dt_format)
    return dt.isoformat('T')
    
def get_dtime_iso(dtime_str: str, z_break: str = 'Z', z_repl: str  = '.000000+00:00'):
    """Breaks the timestamp at z_break and replaces with z_repl if z_break is not empty. """
    # We use central timezone for rancher clusters, so need to convert CST -> UTC
    if z_break: return datetime.fromisoformat(dtime_str.replace(z_break, z_repl))
    return datetime.fromisoformat(dtime_str).astimezone(utc_timezone)

def get_date(timeframe: str = None, future: bool = False):
    if not timeframe: return dtime_now_utc()
    return dtime_parse(timeframe=timeframe, future=future)

def get_dtime_secs(dtime: datetime = None, start: datetime = None, as_cls: bool = False):
    if dtime and start: return (start - dtime).total_seconds()
    if as_cls: return TimeCls((dtime_now() - dtime).total_seconds())
    try: return (dtime_now_utc() - dtime).total_seconds()
    except: return (dtime_now() - dtime).total_seconds()

dtime = get_dtime
dtstr = get_dtime_str
dtsecs = get_dtime_secs
dtnow = dtime_now
dtnow_utc = dtime_now_utc
get_timestamp = dtime_now
get_timestamp_tz = dtime_now_tz
get_timestamp_utc = dtime_now_utc
dtiso = get_dtime_iso

__all__ = [
    'time',
    'datetime',
    'dtime_now',
    'dtime_now_tz',
    'dtime_now_utc',
    'get_timestamp',
    'get_timestamp_tz',
    'get_timestamp_utc',
    'api_timezone',
    'utc_timezone',
    'timezone_format',
    'timer',
    'dtime_parse',
    'dtime_diff',
    'get_dtime',
    'get_dtime_str',
    'get_dtime_iso',
    'get_date',
    'get_dtime_secs',
    'TimeCls',
]


from __future__ import annotations

from lazy.types import *
from .base_imports import *
from .config import FastAPIConfigz, AppConfigz

if _fastapi_available:
    from fastapi import Header, Depends, Body, FastAPI, APIRouter, HTTPException, BackgroundTasks, status
    from fastapi.responses import JSONResponse, PlainTextResponse, HTMLResponse, FileResponse, RedirectResponse, ORJSONResponse, StreamingResponse
    from fastapi import WebSocket, WebSocketDisconnect
else:
    Header, Depends, Body, FastAPI, APIRouter, HTTPException, BackgroundTasks, status = Callable, Callable, Callable, object, object, object, object, object
    JSONResponse, PlainTextResponse, HTMLResponse, FileResponse, RedirectResponse, ORJSONResponse, StreamingResponse = object, object, object, object, object, object, object
    WebSocket, WebSocketDisconnect = object, object

if _starlette_available:
    from starlette.requests import Request
else: Request = object


if TYPE_CHECKING:
    from fastapi import FastAPI, Header, Depends, Body, HTTPException, status, BackgroundTasks
    from fastapi.responses import JSONResponse, PlainTextResponse
    from starlette.requests import Request

def create_fastapi(
        app_name: str = None, 
        title: str = AppConfigz.title, 
        description: str = AppConfigz.description, 
        version: str = AppConfigz.version, 
        include_middleware: bool = FastAPIConfigz.include_middleware, 
        allow_credentials: bool = FastAPIConfigz.allow_credentials, 
        allow_origins: List[str] = FastAPIConfigz.allow_origins, 
        allow_methods: List[str] = FastAPIConfigz.allow_methods, 
        allow_hosts: List[str] = FastAPIConfigz.allow_hosts, 
        allow_headers: List[str] = FastAPIConfigz.allow_headers, 
        auth_config: Dict[str, Any] = None,
        logger: Optional[Any] = None,
        **kwargs
    ) -> Type['FastAPI']:
    """
    This method will dynamically create a new FastAPI App instance using 
    the base Environment Variables in FastAPIConfigz

    In certain use cases (such as submounts), you can leave the app_name blank for the primary app,
    and define the app_name for all subapps.
    """
    if app_name: title += ': ' + app_name
    app_config = AppConfigz()
    app_config.update_config(title = title, description = description, version = version)
    fast_config = FastAPIConfigz()
    fast_config.update_config(include_middleware = include_middleware, allow_credentials = allow_credentials, allow_origins = allow_origins, allow_methods = allow_methods, allow_hosts = allow_hosts, allow_headers = allow_headers)
    new_fastapi_app = fast_config.get_fastapi_app(app_config = app_config, **kwargs)
    if include_middleware: fast_config.update_fastapi_middleware(new_fastapi_app, auth_config)
    if logger: new_fastapi_app.logger = logger
    return new_fastapi_app
    


class FastAPIValidator:
    header: 'Header' = None
    def __init__(self, key: str, alias: str = 'token'):
        self._key = key
        self._alias = alias
        self.header = Header(..., alias=self._alias)

def create_validator(key: str, alias: str = 'token') -> Type[Callable]:
    async def verify_token(token: str = Header(..., alias=alias)):
        if token == key: return True
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid authentication credentials", headers={"WWW-Authenticate": "Bearer"})
    return verify_token


def create_multi_validator(keys: List[str], alias: str = 'token') -> Type[Callable]:
    async def verify_token(token: str = Header(..., alias=alias)):
        if token in keys: return True
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid authentication credentials", headers={"WWW-Authenticate": "Bearer"})
    return verify_token


def create_func_multi_validator_body(func: Callable, keys: List[str], alias: str = 'user_id', data_key: str = None, **funcKwargs) -> Type[Callable]:
    datakey = data_key or alias
    async def verify_data(data: Dict[str, Any] = Body(..., alias=alias)):
        if data.get(datakey) and func(data[datakey], **funcKwargs) in keys: return True
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid authentication credentials", headers={"WWW-Authenticate": "Bearer"})
    return verify_data


def create_func_multi_validator_request(func: Callable, keys: List[str], alias: str = 'user_id', data_key: str = None, reqType: str = 'form', **funcKwargs) -> Type[Callable]:
    datakey = data_key or alias
    async def verify_data(req: 'Request' = Body(..., alias=alias)):
        reqMethod = getattr(req, reqType)
        req = await reqMethod()
        if req.get(datakey) and func(req[datakey], **funcKwargs) in keys: return True
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid authentication credentials", headers={"WWW-Authenticate": "Bearer"})
    return verify_data


## Directly implemented from
## https://fastapi.tiangolo.com/advanced/websockets/

class WebsocketManager:
    def __init__(self):
        _ensure_api_reqs()
        self.active_connections: List[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def send_personal_message(self, message: str, websocket: WebSocket):
        await websocket.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


__all__ = [
    ## FastAPI Base Imports
    'Header', 
    'Depends', 
    'Body', 
    'FastAPI', 
    'APIRouter',
    'HTTPException', 
    'BackgroundTasks', 
    'status',
    ## Responses
    'JSONResponse', 
    'PlainTextResponse', 
    'HTMLResponse', 
    'FileResponse', 
    'RedirectResponse', 
    'ORJSONResponse', 
    'StreamingResponse',
    ## Starlette Request
    'Request',
    ## Websocket
    'WebSocket', 
    'WebSocketDisconnect',
    'WebsocketManager',
    ## Custom Classes / Funcs
    'create_validator',
    'create_multi_validator',
    'create_func_multi_validator_body',
    'create_func_multi_validator_request',
    'AppConfigz',
    'FastAPIConfigz',
    'create_fastapi',
]


from . import static
from .config import CachezConfigz, SqlConfig
from .base import (
    Cache,
    Disk,
    CDisk,
    EmptyDirWarning,
    Timeout,
    UnknownFileWarning,
)
from ._json import OrJSONDisk, JSONDisk
from .persistent import Index, Deque


from __future__ import annotations
import json
from lazy.types import *
from lazy.types.pyd import Json
from lazy.models import Field
from lazy.configz import ConfigCls
from lazy.configz.common import AppBaseConfigCls, PostgresConfigz, RedisConfigz, ElasticsearchConfigz, MysqlConfigz

from logz import get_cls_logger
from .base_imports import httpx, fastapi, FastAPI, _ensure_api_reqs, Lib
if TYPE_CHECKING:
    import httpx
    import fastapi
    from fastapi import FastAPI

get_logger = get_cls_logger('lazy:api')
logger = get_logger()

DefaultHeaders = {
    'Accept': 'application/json',
    'Content-Type': 'application/json'
}

class TimeZoneConfigz(ConfigCls):
    desired: str = 'America/Chicago'
    tz_format: str = '%Y-%m-%dT%H:%M:%SZ'

class HttpConfigz(ConfigCls):
    base_url: str = ""
    timeout: float = 30.0
    keep_alive: int = 50
    max_connect: int = 200
    default_headers: Json = json.dumps(DefaultHeaders)
    module_name: str = 'lazy'
    
    @property
    def httpx_timeout(self):
        return httpx.Timeout(self.timeout, connect=self.timeout)

    @property
    def httpx_limits(self):
        return httpx.Limits(max_keepalive_connections=self.keep_alive, max_connections=self.max_connect)
    
    @property
    def httpx_config(self):
        return {'timeout': self.httpx_timeout, 'limits': self.httpx_limits, 'headers': self.default_headers}


class AsyncHttpConfigz(HttpConfigz):
    pass

class AppConfigz(AppBaseConfigCls):
    pass

## Basically a copy paste from configz.common but more explicit

class FastAPIConfigz(ConfigCls):
    root_path: Optional[str] = None
    api_prefix: Optional[str] = ''
    xapi_prefix: Optional[str] = '/xapi'
    graphql_prefix: Optional[str] = '/graphql'
    openapi_prefix: Optional[str] = ''
    openapi_url: Optional[str] = '/openapi.json'
    docs_url: Optional[str] = '/docs'
    redoc_url: Optional[str] = '/redoc'
    version: Optional[str] = '0.0.1'
    root_path_in_servers: Optional[bool] = False
    include_middleware: Optional[bool] = True
    allow_credentials: Optional[bool] = True
    allow_hosts: Optional[List[str]] = []
    allow_origins: Optional[List[str]] = ["*"]
    allow_methods: Optional[List[str]] = ["*"]
    allow_headers: Optional[List[str]] = ["*"]
    app_configz: Optional[Json]

    class Config:
        env_prefix = "FASTAPI_"

    @property
    def app_config(self):
        if not self.app_configz:
            self.app_configz = AppConfigz()
        elif isinstance(self.app_configz, dict):
            configz = AppConfigz()
            configz.update_config(**self.app_configz)
            self.app_configz = configz
        return self.app_configz
    
    @property
    def app_prefix(self) -> str:
        """
        root_path takes precendence over api_prefix
        """
        return self.root_path or self.api_prefix

    @property
    def openapi_path(self):
        """ 
        Prefixes openapi_url with openapi_prefix/app_prefix 
        with openapi_prefix taking precedence
        """
        p: str = (self.openapi_prefix or self.app_prefix) + self.openapi_url
        ## Fix any double slashes when combining path
        p = p.replace('//', '/')
        return p
    
    def get_openapi_prefix(self):
        """ 
        since openapi_url should either be a fixed path, or have prefix
        if openapi_prefix is present, this should be None
        """
        if self.openapi_prefix in self.openapi_path: return ''
        return self.openapi_prefix

    def get_fastapi_config(self, app_config: Union[AppConfigz, Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Returns Dict for FastAPI(**config)
        """
        if app_config:
            if isinstance(app_config, type(AppConfigz)):
                self.app_configz = app_config
            else:
                self.app_config.update_config(**app_config) # self.app_config = self.app_config.update_config(**app_config)
        return {
            'title': self.app_config.title,
            'version': self.app_config.version or self.version,
            'root_path': self.app_prefix,
            'openapi_prefix': self.get_openapi_prefix(),
            'openapi_url': self.openapi_path,
            'root_path_in_servers': self.root_path_in_servers,
        }
            
    def get_fastapi_app(self, app_config: Union[AppConfigz, Dict[str, Any]] = None, **kwargs) -> 'FastAPI':
        """
        Initializes a new FastAPI App using Config Settings

        returns fastapi.FastAPI
        """
        _ensure_api_reqs()
        config = self.get_fastapi_config(app_config)
        if kwargs: config.update(kwargs)
        return FastAPI(**config)
    
    def update_fastapi_middleware(self, app: 'FastAPI', auth_config: Dict[str, Any] = None):
        """
        Updates an existing FastAPI instance with middlewares
        Assumes fastapi is installed since it would be required to pass the app

        returns the same app
        """
        _ensure_api_reqs()
        if auth_config: self.update_config(**auth_config) #self = self.update_config(**auth_config)
        from starlette.middleware.cors import CORSMiddleware
        app.add_middleware(
            CORSMiddleware,
            allow_origins=self.allow_origins,
            allow_credentials=self.allow_credentials,
            allow_methods=self.allow_methods,
            allow_headers=self.allow_headers,
        )
        if self.allow_hosts:
            from fastapi.middleware.trustedhost import TrustedHostMiddleware
            app.add_middleware(TrustedHostMiddleware, allowed_hosts=self.allow_hosts)
        return app


__all__ = [
    'logger',
    'DefaultHeaders',
    'TimeZoneConfigz',

    'HttpConfigz',
    'AsyncHttpConfigz',
    
    'AppConfigz',
    'FastAPIConfigz',

    'PostgresConfigz', 
    'RedisConfigz', 
    'ElasticsearchConfigz', 
    'MysqlConfigz'

]



from __future__ import annotations

from lazy.types import *
from lazy.models import BaseCls

from .config import *
from .types import *
from .utils import convert_to_cls
from .base_imports import _httpx_available, _ensure_api_reqs

if _httpx_available:
    from httpx import Client as _Client
    from httpx import AsyncClient as _AsyncClient
    from httpx import Response as HttpResponse
else:
    _Client, _AsyncClient, HttpResponse = object, object, object


class Client:
    _web: _Client = None
    _async: _AsyncClient = None 
    
    @classmethod
    def create_client(cls, base_url: str = "", config: Dict[str, Any] = None, **kwargs) -> Type[_Client]:
        """Creates a Sync httpx Client"""
        _ensure_api_reqs()
        configz = HttpConfigz()
        if config: configz.update_config(**config)
        client_config = configz.httpx_config
        if 'headers' in kwargs:
            headers = kwargs.pop('headers')
            if headers: client_config['headers'] = headers
        return _Client(base_url = base_url, **client_config, **kwargs)
    
    @classmethod
    def create_async_client(cls, base_url: str = "", config: Dict[str, Any] = None, **kwargs) -> Type[_AsyncClient]:
        """ Creates an async httpx Client"""
        _ensure_api_reqs()
        configz = AsyncHttpConfigz()
        if config: configz.update_config(**config)
        client_config = configz.httpx_config
        if 'headers' in kwargs:
            headers = kwargs.pop('headers')
            if headers: client_config['headers'] = headers
        return _AsyncClient(base_url = base_url, **client_config, **kwargs)

    @classproperty
    def client(cls) -> Type[_Client]:
        if not cls._web: cls._web = cls.create_client()
        return cls._web
    
    @classproperty
    def async_client(cls) -> Type[_AsyncClient]:
        if not cls._async: cls._async = cls.create_async_client()
        return cls._async



class ApiClient:
    def __init__(self, base_url: str = HttpConfigz.base_url or AsyncHttpConfigz.base_url, headers: DictAny = {}, config: DictAny = None, async_config: DictAny = None, module_name: str = HttpConfigz.module_name or AsyncHttpConfigz.module_name, default_resp: bool = False, **kwargs):
        _ensure_api_reqs()
        self.base_url = ""
        self.headers = {}
        self.config = None
        self.async_config = None
        self._module_name = None
        self._kwargs = {}
        self._web = None
        self._async = None
        self._default_mode = False
        self.set_configs(base_url = base_url, headers = headers, config = config, async_config = async_config, module_name = module_name, default_resp = default_resp, **kwargs)

    def set_configs(self, base_url: str = HttpConfigz.base_url or AsyncHttpConfigz.base_url, headers: DictAny = {}, config: DictAny = None, async_config: DictAny = None, module_name: str = HttpConfigz.module_name or AsyncHttpConfigz.module_name, default_resp: bool = False,  **kwargs):
        self.base_url = base_url or self.base_url
        self.headers = headers or self.headers
        self.config = config or self.config
        self.async_config = async_config or self.async_config
        self._module_name = module_name or self._module_name
        self._default_mode = default_resp or self._default_mode
        self._kwargs = kwargs or self._kwargs

    def reset_clients(self, base_url: str = HttpConfigz.base_url or AsyncHttpConfigz.base_url, headers: DictAny = {}, config: DictAny = None, async_config: DictAny = None, module_name: str = HttpConfigz.module_name or AsyncHttpConfigz.module_name, default_resp: bool = False, **kwargs):
        self.set_configs(base_url = base_url, headers = headers, config = config, async_config = async_config, module_name = module_name, default_resp = default_resp, **kwargs)
        self._web = None
        self._async = None
    
    @property
    def client(self):
        if not self._web: self._web = Client.create_client(base_url=self.base_url, config=self.config, headers=self.headers, **self._kwargs)
        return self._web
    
    @property
    def aclient(self):
        if not self._async: self._async = Client.create_async_client(base_url=self.base_url, config=self.async_config, headers=self.headers, **self._kwargs)
        return self._async
    

    #############################################################################
    #                             Base REST APIs                                #
    #############################################################################
    
    def delete(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.delete(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'delete')

    def get(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.get(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'get')

    def head(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.head(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'head')

    def patch(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.patch(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'patch')

    def put(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.put(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'put')
    
    def post(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = self.client.post(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'sync', method = 'post')

    #############################################################################
    #                          Async REST Methods                               #
    #############################################################################
    
    async def async_delete(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.delete(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'delete')

    async def async_get(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.get(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'get')
    
    async def async_head(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.head(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'head')

    async def async_patch(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.patch(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'patch')

    async def async_put(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.put(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'put')
    
    async def async_post(self, path: str, **kwargs) -> Union[Response, HttpResponse]:
        resp = await self.aclient.post(url=path, **kwargs)
        if self._default_mode: return resp
        return Response(resp = resp, client_type = 'async', method = 'post')


    #############################################################################
    #                       Supplementary Helpful Callers                       #
    #############################################################################

    def ping(self, path: str, max_status_code: int = 300, min_status_code: int = None, **kwargs) -> bool:
        """ Returns a bool of whether response code is great/within range/less than an int
            Can be used as a health check """
        res = self.get(url=path, **kwargs)
        if min_status_code and max_status_code:
            return bool(res.status_code in range(min_status_code, max_status_code))
        if min_status_code:
            return bool(res.status_code > min_status_code)
        return bool(res.status_code < max_status_code)
    
    def get_data(self, path: str, key: str = 'data', **kwargs) -> DataType:
        """ Expects to get data in JSON. If does not get the key, returns None. """
        resp = self.get(url=path, **kwargs)
        return resp.data.get(key, None)
    
    def get_lazycls(self, path: str, key: str = 'data', **kwargs) -> Type[BaseCls]:
        """
        Expects to get data in JSON. If does not get the key, returns None.
        Returns the data from a GET request to Path as a LazyCls
        """
        data = self.get_data(path=path, key=key, **kwargs)
        if not data: return None
        return convert_to_cls(resp=data, module_name=self._module_name, base_key=key)


    #############################################################################
    #                  Async Supplementary Helpful Callers                      #
    #############################################################################

    async def async_ping(self, path: str, max_status_code: int = 300, min_status_code: int = None, **kwargs) -> bool:
        """ Returns a bool of whether response code is great/within range/less than an int
            Can be used as a health check """
        res = await self.async_get(url=path, **kwargs)
        if min_status_code and max_status_code:
            return bool(res.status_code in range(min_status_code, max_status_code))
        if min_status_code:
            return bool(res.status_code > min_status_code)
        return bool(res.status_code < max_status_code)
    
    async def async_get_data(self, path: str, key: str = 'data', **kwargs) -> DataType:
        """ Expects to get data in JSON. If does not get the key, returns None. """
        resp = await self.async_get(url=path, **kwargs)
        return resp.data.get(key, None)
    
    async def async_get_lazycls(self, path: str, key: str = 'data', **kwargs) -> Type[BaseCls]:
        """
        Expects to get data in JSON. If does not get the key, returns None.
        Returns the data from a GET request to Path as a LazyCls
        """
        data = await self.async_get_data(path=path, key=key, **kwargs)
        if not data: return None
        return convert_to_cls(resp=data, module_name=self._module_name, base_key=key)
    

        
APIClient = ApiClient

__all__ = [
    'Client',
    'HttpResponse',
    'ApiClient',
    'APIClient',
    '_Client',
    '_AsyncClient'
]



from lazy.models import BaseCls, create_lazycls
from lazy.types import *

def convert_to_cls(resp: DictAny, module_name: str = 'lazy', base_key: str = 'api') -> List[Type[BaseCls]]:
    for key, vals in resp.items():
        mod_key = f'{base_key}{key}'
        if isinstance(vals, list): vals = [create_lazycls(mod_key, v, modulename=module_name) for v in vals]
        else: vals = create_lazycls(mod_key, vals, modulename=module_name)
        resp[key] = vals
    return resp





"""
Borrowed from

https://github.com/slingamn/mureq/blob/master/mureq.py

Why?
In short: performance (memory consumption), security (resilience to supply-chain attacks), and simplicity.

Performance
python-requests is extremely memory-hungry, mainly due to large transitive dependencies like chardet that are not needed

mureq is a replacement for python-requests, intended to be vendored
in-tree by Linux systems software and other lightweight applications.

mureq is copyright 2021 by its contributors and is released under the
0BSD ("zero-clause BSD") license.
"""

import contextlib
import io
import os.path
import socket
import ssl
import sys
import urllib.parse
from http.client import HTTPConnection, HTTPSConnection, HTTPMessage, HTTPException
from lazy.static.web import DefaultHeaders


__all__ = ['HTTPException', 'TooManyRedirects', 'Response', 'yield_response', 'request',  'get', 'post', 'head', 'put', 'patch', 'delete']

DEFAULT_TIMEOUT = 15.0
MAX_REDIRECTS = 2
DEFAULT_HEADERS = DefaultHeaders

# e.g. "Python 3.8.10"
DEFAULT_UA = "Python " + sys.version.split()[0]


def request(method, url, *, read_limit=None, **kwargs):
    """request performs an HTTP request and reads the entire response body.
    :param str method: HTTP method to request (e.g. 'GET', 'POST')
    :param str url: URL to request
    :param read_limit: maximum number of bytes to read from the body, or None for no limit
    :type read_limit: int or None
    :param **kwargs: optional arguments defined by yield_response
    :return: Response object
    :rtype: Response
    :raises: HTTPException
    """
    with yield_response(method, url, **kwargs) as response:
        try:
            body = response.read(read_limit)
        except IOError as e:
            if isinstance(e, HTTPException): raise
            else: raise HTTPException(str(e)) from e
        return Response(response.url, response.status, _prepare_incoming_headers(response.headers), body)

def get(url, **kwargs):
    """get performs a HTTP GET request."""
    return request('GET', url=url, **kwargs)

def post(url, body=None, **kwargs):
    """post performs a HTTP POST request."""
    return request('POST', url=url, body=body, **kwargs)

def head(url, **kwargs):
    """head performs a HTTP HEAD request."""
    return request('HEAD', url=url, **kwargs)

def put(url, body=None, **kwargs):
    """put performs a HTTP PUT request."""
    return request('PUT', url=url, body=body, **kwargs)

def patch(url, body=None, **kwargs):
    """patch performs a HTTP PATCH request."""
    return request('PATCH', url=url, body=body, **kwargs)

def delete(url, **kwargs):
    """delete performs a HTTP DELETE request."""
    return request('DELETE', url=url, **kwargs)

@contextlib.contextmanager
def yield_response(method, url, *, unix_socket=None, timeout=DEFAULT_TIMEOUT, headers=None,
        params=None, body=None, form=None, json=None, verify=True, source_address=None,
        max_redirects=MAX_REDIRECTS, ssl_context=None):
    """yield_response is a low-level API that exposes the actual
    http.client.HTTPResponse via a contextmanager.
    Note that unlike mureq.Response, http.client.HTTPResponse does not
    automatically canonicalize multiple appearances of the same header by
    joining them together with a comma delimiter. To retrieve canonicalized
    headers from the response, use response.getheader():
    https://docs.python.org/3/library/http.client.html#http.client.HTTPResponse.getheader
    :param str method: HTTP method to request (e.g. 'GET', 'POST')
    :param str url: URL to request
    :param unix_socket: path to Unix domain socket to query, or None for a normal TCP request
    :type unix_socket: str or None
    :param timeout: timeout in seconds, or None for no timeout (default: 15 seconds)
    :type timeout: float or None
    :param headers: HTTP headers as a mapping or list of key-value pairs
    :param params: parameters to be URL-encoded and added to the query string, as a mapping or list of key-value pairs
    :param body: payload body of the request
    :type body: bytes or None
    :param form: parameters to be form-encoded and sent as the payload body, as a mapping or list of key-value pairs
    :param json: serialized JSON data to be sent as the payload body
    :type json: str or bytes or None
    :param bool verify: whether to verify TLS certificates (default: True)
    :param source_address: source address to bind to for TCP
    :type source_address: str or tuple(str, int) or None
    :param max_redirects: maximum number of redirects to follow, or None (the default) for no redirection
    :type max_redirects: int or None
    :param ssl_context: TLS config to control certificate validation, or None for default behavior
    :type ssl_context: ssl.SSLContext or None
    :return: http.client.HTTPResponse, yielded as context manager
    :rtype: http.client.HTTPResponse
    :raises: HTTPException
    """
    method = method.upper()
    headers = _prepare_outgoing_headers(headers)
    enc_params = _prepare_params(params)
    body = _prepare_body(body, form, json, headers)

    visited_urls = []

    while max_redirects is None or len(visited_urls) <= max_redirects:
        url, conn, path = _prepare_request(method, url, enc_params=enc_params, timeout=timeout, unix_socket=unix_socket, verify=verify, source_address=source_address, ssl_context=ssl_context)
        enc_params = '' # don't reappend enc_params if we get redirected
        visited_urls.append(url)
        try:
            try:
                conn.request(method, path, headers=headers, body=body)
                response = conn.getresponse()
            except IOError as e:
                if isinstance(e, HTTPException):
                    raise
                else:
                    # wrap any IOError that is not already an HTTPException
                    # in HTTPException, exposing a uniform API for remote errors
                    raise HTTPException(str(e)) from e
            redirect_url = _check_redirect(url, response.status, response.headers)
            if max_redirects is None or redirect_url is None:
                response.url = url # https://bugs.python.org/issue42062
                yield response
                return
            else:
                url = redirect_url
                if response.status == 303:
                    # 303 See Other: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/303
                    method = 'GET'
        finally:
            conn.close()

    raise TooManyRedirects(visited_urls)


class Response:
    """Response contains a completely consumed HTTP response.
    :ivar str url: the retrieved URL, indicating whether a redirection occurred
    :ivar int status_code: the HTTP status code
    :ivar http.client.HTTPMessage headers: the HTTP headers
    :ivar bytes body: the payload body of the response
    """

    __slots__ = ('url', 'status_code', 'headers', 'body')

    def __init__(self, url, status_code, headers, body):
        self.url, self.status_code, self.headers, self.body = url, status_code, headers, body

    def __repr__(self):
        return "Response(status_code=%d)" % (self.status_code,)

    @property
    def ok(self):
        """ok returns whether the response had a successful status code
        (anything other than a 40x or 50x)."""
        return not (400 <= self.status_code and self.status_code < 600)

    @property
    def content(self):
        """content returns the response body (the `body` member). This is an
        alias for compatibility with requests.Response."""
        return self.body

    def _debugstr(self):
        buf = io.StringIO()
        print("HTTP", self.status_code, file=buf)
        for k, v in self.headers.items():
            print("%s: %s" % (k, v), file=buf)
        print(file=buf)
        try:
            print(self.body.decode('utf-8'), file=buf)
        except UnicodeDecodeError:
            print("<%d bytes binary data>" % (len(self.body),), file=buf)
        return buf.getvalue()


class TooManyRedirects(HTTPException):
    """TooManyRedirects is raised when automatic following of redirects was
    enabled, but the server redirected too many times without completing."""
    pass


# end public API, begin internal implementation details

_JSON_CONTENTTYPE = 'application/json'
_FORM_CONTENTTYPE = 'application/x-www-form-urlencoded'


class UnixHTTPConnection(HTTPConnection):
    """UnixHTTPConnection is a subclass of HTTPConnection that connects to a
    Unix domain stream socket instead of a TCP address.
    """

    def __init__(self, path, timeout=DEFAULT_TIMEOUT):
        super(UnixHTTPConnection, self).__init__('localhost', timeout=timeout)
        self._unix_path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.settimeout(self.timeout)
            sock.connect(self._unix_path)
        except:
            sock.close()
            raise
        self.sock = sock

def _check_redirect(url, status, response_headers):
    """Return the URL to redirect to, or None for no redirection."""
    if status not in (301, 302, 303, 307, 308): return None

    location = response_headers.get('Location')
    if not location: return None

    parsed_location = urllib.parse.urlparse(location)
    if parsed_location.scheme:
        # absolute URL
        return location

    old_url = urllib.parse.urlparse(url)
    if location.startswith('/'):
        # absolute path on old hostname
        return urllib.parse.urlunparse((old_url.scheme, old_url.netloc, parsed_location.path, parsed_location.params, parsed_location.query, parsed_location.fragment))

    # relative path on old hostname
    old_dir, _old_file = os.path.split(old_url.path)
    new_path = os.path.join(old_dir, location)
    return urllib.parse.urlunparse((old_url.scheme, old_url.netloc, new_path, parsed_location.params, parsed_location.query, parsed_location.fragment))

def _prepare_outgoing_headers(headers):
    if headers is None:
        headers = HTTPMessage()
    elif not isinstance(headers, HTTPMessage):
        new_headers = HTTPMessage()
        if hasattr(headers, 'items'):
            iterator = headers.items()
        else:
            iterator = iter(headers)
        for k, v in iterator:
            new_headers[k] = v
        headers = new_headers
    _setdefault_header(headers, 'User-Agent', DEFAULT_UA)
    return headers

# XXX join multi-headers together so that get(), __getitem__(),
# etc. behave intuitively, then stuff them back in an HTTPMessage.
def _prepare_incoming_headers(headers):
    headers_dict = {}
    for k, v in headers.items():
        headers_dict.setdefault(k, []).append(v)
    result = HTTPMessage()
    # note that iterating over headers_dict preserves the original
    # insertion order in all versions since Python 3.6:
    for k, vlist in headers_dict.items():
        result[k] = ','.join(vlist)
    return result

def _setdefault_header(headers, name, value):
    if name not in headers:
        headers[name] = value

def _prepare_body(body, form, json, headers):
    if body is not None:
        if not isinstance(body, bytes): raise TypeError('body must be bytes or None', type(body))
        return body

    if json is not None:
        if isinstance(json, bytes):
            _setdefault_header(headers, 'Content-Type', _JSON_CONTENTTYPE)
            return json
        elif isinstance(json, str):
            _setdefault_header(headers, 'Content-Type', _JSON_CONTENTTYPE)
            return json.encode('utf-8')
        else:
            raise TypeError('json must be str, bytes, or None', type(json))

    if form is not None:
        _setdefault_header(headers, 'Content-Type', _FORM_CONTENTTYPE)
        return urllib.parse.urlencode(form, doseq=True)

    return None

def _prepare_params(params):
    if params is None:
        return ''
    return urllib.parse.urlencode(params, doseq=True)

def _prepare_request(method, url, *, enc_params='', timeout=DEFAULT_TIMEOUT, source_address=None, unix_socket=None, verify=True, ssl_context=None):
    """Parses the URL, returns the path and the right HTTPConnection subclass."""
    parsed_url = urllib.parse.urlparse(url)

    is_unix = (unix_socket is not None)
    scheme = parsed_url.scheme.lower()
    if scheme.endswith('+unix'):
        scheme = scheme[:-5]
        is_unix = True
        if scheme == 'https': raise ValueError("https+unix is not implemented")

    if scheme not in ('http', 'https'): raise ValueError("unrecognized scheme", scheme)

    is_https = (scheme == 'https')
    host = parsed_url.hostname
    port = 443 if is_https else 80
    if parsed_url.port: port = parsed_url.port

    if is_unix and unix_socket is None: unix_socket = urllib.parse.unquote(parsed_url.netloc)

    path = parsed_url.path
    if parsed_url.query:
        if enc_params: path = '%s?%s&%s' % (path, parsed_url.query, enc_params)
        else: path = '%s?%s' % (path, parsed_url.query)
    else:
        if enc_params: path = '%s?%s' % (path, enc_params)
        else: pass # just parsed_url.path in this case

    if isinstance(source_address, str): source_address = (source_address, 0)

    if is_unix: conn = UnixHTTPConnection(unix_socket, timeout=timeout)
    elif is_https:
        if ssl_context is None:
            ssl_context = ssl.create_default_context()
            if not verify:
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE
        conn = HTTPSConnection(host, port, source_address=source_address, timeout=timeout, context=ssl_context)
    
    else: conn = HTTPConnection(host, port, source_address=source_address, timeout=timeout)

    munged_url = urllib.parse.urlunparse((parsed_url.scheme, parsed_url.netloc, path, parsed_url.params, '', parsed_url.fragment))
    return munged_url, conn, path


"""
Performance Utilz
"""
import time
import anyio
import multiprocessing
from functools import wraps, partial

from typing import Callable, Dict, List, Any, Union, Iterable, Generator, Optional


from .utils import *
from ..helpers import get_logger

try:
    import trio
    _async_backend = 'trio'

except ImportError:
    _async_backend = 'asyncio'

"""
Some Base Variables that don't rely on other 
modules from lazy to ensure no circular dependencies
"""


logger = get_logger('lazy:ops')

NUM_CPUS = multiprocessing.cpu_count()
MAX_THREADS = round(NUM_CPUS * 0.7)
MAX_PROCESSES = round(NUM_CPUS * 0.5)
MAX_WORKERS = round(NUM_CPUS * 0.5)


def get_mp():
    """
    Gets the correct Multiprocessing Library for win/mac/linux
    """
    return (multiprocessing.get_context('fork') if hasattr(multiprocessing, 'get_context') else multiprocessing)



def ProcessPool(process: Callable, workers: int = MAX_WORKERS, has_args: bool = False, has_kwargs: bool = False, verbose: bool = False, timed: bool = False, show_progress: bool = False):
    # sourcery no-metrics
    """
    Creates a wrapper function that leverages
    multiprocessing to pass all items yielded by this
    function to the [process]

    :process                = Callable
    :workers                = num of max workers
    :has_args [False]       = if the returned result is a list, will pass as *args if True
    :has_kwargs [False]     = if the returned result is a dict, will pass as **kwargs if True
    :verbose [False]        = Will log start/stop
    :timed  [False]         = Measures the time of function call.
    :show_progress [False]  = Displays a Progress bar (tqdm) if available
    """

    def put_to_queue(input_queue: multiprocessing.Queue, wrapped_func: Union[List[Any], Dict[Any, Any], Callable, Generator, Iterable, Any]):
        if isinstance(wrapped_func, Callable):
            for item in wrapped_func():
                input_queue.put(item)

        else:
            for item in wrapped_func:
                input_queue.put(item)

        for _ in range(workers):
            input_queue.put(None)
    
    def process_item(input_queue: multiprocessing.Queue, output_queue: multiprocessing.Queue):
        while True:
            item = input_queue.get()
            if item is None:
                output_queue.put(item)
                break
            if has_kwargs and isinstance(item, dict):
                output_queue.put(process(**item))
            elif has_args and isinstance(item, list):
                output_queue.put(process(*item))
            else:
                output_queue.put(process(item))

    def wrapped_generator(func):
        
        @wraps(func)
        def process_pool(*args, **kwargs):
            mp = get_mp()
            input_queue = mp.Queue(maxsize=workers)
            output_queue = mp.Queue(maxsize=workers)

            wrapped_func = partial(func, *args, **kwargs)
            _generator_pool = mp.Pool(1, initializer=put_to_queue, initargs=(input_queue, wrapped_func))
            _process_pool = mp.Pool(workers, initializer=process_item, initargs=(input_queue, output_queue))
            if verbose: logger.info(f'Starting ProcessPool for Func: {func.__name__} with {workers} workers. Timed = {timed}')
            ts = time.perf_counter() if timed else None
            pbar: Optional['trange'] = wrap_tqdm_iterable(wrapped_func, desc = f'ProcessPool: {func.__name__}') if show_progress and _tqdm_enabled else None
            workers_done = 0
            while True:
                item = output_queue.get()
                if item is None:
                    workers_done += 1
                    if workers_done == workers:
                        break
                else:
                    if pbar: pbar.update()
                    yield item
                    
            _generator_pool.close()
            _generator_pool.join()
            _process_pool.close()
            _process_pool.join()

            if timed:
                done_time = time.perf_counter() - ts
                if verbose: logger.info(f'Completed ProcessPool for Func: {func.__name__} with {workers} workers in {done_time:.2f} secs')
                return done_time
            if verbose: logger.info(f'Completed ProcessPool for Func: {func.__name__} with {workers} workers')
        
        return process_pool
        
    return wrapped_generator


    
def AsyncTaskGroup(process: Callable, verbose: bool = False, timed: bool = False, show_progress: bool = False):
    """
    Creates an async wrapper function that leverages
    anyio.create_task_group to pass all items yielded by this
    function to the [process]

    :process                = Callable. Will convert to async function.
    :verbose [False]        = Will log start/stop
    :timed  [False]         = Measures the time of function call.
    :show_progress [False]  = Displays a Progress bar (tqdm) if available / Not used atm
    """

    async def task_group_process(wrapped_func):
        process_func = process
        if not iscoroutinefunction(process_func):
            process_func = AsyncFunction(process_func)

        async with anyio.create_task_group() as tg:
            for item in await wrapped_func():
                tg.start_soon(process_func, item)

    def wrapped_task_group(func):
        @wraps(func)
        def async_task_group(*args, **kwargs):
            if not iscoroutinefunction(func): 
                wrapped_func = AsyncFunction(func)
            else: 
                wrapped_func = partial(func, *args, **kwargs)

            if verbose: logger.info(f'Starting AsyncTaskGroup for Func: {func.__name__}. Timed = {timed}')
            ts = time.perf_counter() if timed else None
            anyio.run(task_group_process, wrapped_func, backend=_async_backend)
            if timed:
                done_time = time.perf_counter() - ts
                if verbose: logger.info(f'Completed AsyncTaskGroup for Func: {func.__name__} in {done_time:.2f} secs')
                return done_time
            if verbose: logger.info(f'Completed AsyncTaskGroup for Func: {func.__name__}')

        return async_task_group
        
    return wrapped_task_group




__all__ = (
    'full_name',
    'WindowsExceptionError',
    'Constant',
    'ENOVAL',
    'UNKNOWN',
    'MODE_NONE',
    'MODE_RAW',
    'MODE_BINARY',
    'MODE_TEXT',
    'MODE_PICKLE',
    'METADATA',
)

def full_name(func):
    "Return full name of `func` by adding the module and function name."
    return func.__module__ + '.' + func.__qualname__


try: 
    class WindowsExceptionError(WindowsError):
        pass
except:
    class WindowsExceptionError(Exception):
        "Windows error place-holder on platforms without support."


class Constant(tuple):
    "Pretty display of immutable constant."

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return '%s' % self[0]

ENOVAL = Constant('ENOVAL')
UNKNOWN = Constant('UNKNOWN')

MODE_NONE = 0
MODE_RAW = 1
MODE_BINARY = 2
MODE_TEXT = 3
MODE_PICKLE = 4

METADATA = {
    u'count': 0,
    u'size': 0,
    u'hits': 0,
    u'misses': 0,
}



"""
Consolidating lazyapi into this.
"""
from . import types

from .config import (
    AppConfigz, FastAPIConfigz,
    TimeZoneConfigz, 
    HttpConfigz, AsyncHttpConfigz,
    PostgresConfigz, RedisConfigz, ElasticsearchConfigz, MysqlConfigz
)
from .types import *
from .client import *
from .fast import *
from .backends import RedisBackend


from . import helpers
from . import procs

from . import _async

from .helpers import get_logger, get_cls_logger, timer
from . import wrapz

logger = get_logger('lazy')
_get_logger = get_cls_logger('lazy')


import time

from lazy.types import *
from lazy.libz import Lib
from lazy.io import get_path
from .config import logger
from .config import RedisConfigz

from lazy.utils.procs import exec_daemon

import time

## port of internal lib used to start services
try: 
    from redis import Redis as RedisClient
except ImportError: 
    RedisClient: object = None

_HOST_OS = None
_REDIS_ENABLED = False
_REDIS_PY = False
_REDIS_INSTALL = None

def _set_host_os():
    global _HOST_OS
    if _HOST_OS is not None: return
    import platform
    _HOST_OS = platform.system().lower()

def _set_redis_install():
    global _REDIS_INSTALL
    if _REDIS_INSTALL is not None: return
    _set_host_os()
    if _HOST_OS == 'darwin':
        _REDIS_INSTALL = 'redis'
    elif _HOST_OS == 'windows':
        _REDIS_INSTALL = 'redis-64'
    else:
        _REDIS_INSTALL = 'redis-server'

def _get_redispy():
    global RedisClient, _REDIS_PY
    if _REDIS_PY and RedisClient: return
    if RedisClient is None:
        RedisClient = Lib['redis:Redis']
    _REDIS_PY = True


def require_redis():
    global RedisClient, _REDIS_ENABLED
    if _REDIS_ENABLED: return
    if not Lib.is_exec_available('redis-server'):
        _set_redis_install()
        Lib.install_binary(_REDIS_INSTALL)
    _get_redispy()    
    #if RedisClient is None: raise ImportError('Redis')
    #global RedisClient
    #if RedisClient is not None: return


class RedisBackend(RedisConfigz):
    cachedir: str = None
    sentinel: bool = False
    sentinel_master: str = None
    fallback_enabled: bool = True
    local_redis: bool = True
    is_deployment: bool = False
    startup_delay: int = 7
    startup_attempts: int = 10

    class Config:
        env_prefix = "REDIS_"

    def wait_for_redis_started(self):
        r = RedisClient(host=self.host, port=self.port, db=self.database, password=self.password, socket_timeout = 5.0, socket_connect_timeout = 5.0, socket_keepalive = False)
        attempts, success = 0, False
        while attempts < self.startup_attempts:
            try:
                r.ping()
                success = True
                break            
            except:
                time.sleep(self.startup_delay)
        if not success: raise Exception
    
    def check_redis_connection(self):
        """Doing a quick ping and then timeout. If does not connect in 5 secs, then is not fast enough."""
        r = RedisClient(host=self.host, port=self.port, db=self.database, password=self.password, socket_timeout = 5.0, socket_connect_timeout = 5.0, socket_keepalive = False)
        try: return bool(r.ping())
        except: return False

    def set_to_default_redis_config(self):
        """ Changes Redis cfg to default"""
        self.host = '127.0.0.1'
        self.port = 6379
        self.password = None

    def get_cachedir(self):
        d = get_path(self.cachedir) if self.cachedir else Lib.get_cwd('db', string=False)
        d.mkdir(parents=True)
        return d.as_posix()

    def start_redis_local(self):
        """ Ensures that local redis is running if local_redis = true or manually called if not able to connect to external redis"""
        # Doing System Platform Check
        require_redis()
        rstatus = Lib.run_cmd('redis-cli ping', raise_error=False)
        if not rstatus:
            if self.is_deployment: exec_daemon('redis-server', cwd=self.get_cachedir(), set_proc_uid=False)
            else: exec_daemon('redis-server', cwd=self.get_cachedir())
            self.wait_for_redis_started()
        self.set_to_default_redis_config()

    def ensure_redis(self):
        """If external redis, checks connect.
            falls back to creating internal redis.
        """
        if not self.local_redis and self.check_redis_connection(): return logger.info('External Redis Connection Successful')
        if not self.fallback_enabled: raise RuntimeError('Unable to Establish Redis connection')
        if self.local_redis:
            logger.info('Setting up Local Redis Connection')
            self.start_redis_local()
            #if self.check_redis_connection(): 
            return logger.info('Local Redis Connection Successful')
        logger.info('Unable to Establish Redis Connection')
        raise

    @property
    def redis_config(self):
        return dict(
            host = self.host,
            port = self.port,
            password = self.password,
            database = self.database,
            sentinel = self.sentinel,
            sentinel_master = self.sentinel_master
        )
    
